/* Copyright (C) 2003 Juan Lang
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License Version 2 or later
 * (the "GPL").
 *
 * This program implements the DHCPINFORM portion of DHCP (see RFC2131) to get
 * DHCP options.  It prints the data from the DHCP server in ASCII form to
 * stdout.
 * The program's intended to play nice, therefore:
 * - I'm careful about length checking to avoid buffer overruns
 * - since I expect I need to be root under UNIX to bind to the DHCP client
 *   port, I use seteuid() to hold root's euid as little as possible
 * - I allow you to specify which interface to talk through and which DHCP
 *   server you want to try, so you can be nice to your neighbors and avoid
 *   unnecessary broadcasts
 * - I've separated the things I know are platform-specific into functions or
 *   macros and marked them with <platform>-specific
 * See TODO's for stuff I think isn't quite there yet.
 */
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include "iptypes.h"
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_NET_IF_H
#include <net/if.h>
#endif
#ifdef HAVE_NET_IF_ARP_H
#include <net/if_arp.h>
#endif
#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#include "bool.h"
#include "ifdetect.h"
#include "timedelta.h"
#include "iputil.h"
#include "minmax.h"
#include "strcompat.h"

#ifndef ARPHRD_ETHER
#define ARPHRD_ETHER 1
#endif

#define MAX_OPTION 255
/* not enforced, just used in pretty-printing option code */
#define MAX_OPTION_NAME 80

/* non-const globals */
static Bool gVerbose = 0;
static Bool gQuiet = 0;
static Bool gHelp = 0;
static Bool gPrintOptions = 0;
static u_int gXID;
static u_char gMessageType = 0;
static u_char gOverload = 0;
static u_char gRequestedOptions[MAX_OPTION + 1];
static u_int gNumRequestedOptions = 0;
static struct InterfaceInfo *gInterfaces = NULL;
static u_int gNumDestinations = 0;

/* some DHCP-related constants */

/* offsets of things */
static const u_int OpOffset = 0;
static const u_int HTypeOffset = 1;
static const u_int HLenOffset = 2;
static const u_int HOpsOffset = 3;
static const u_int XidOffset = 4;
static const u_int SecsOffset = 8;
static const u_int FlagsOffset = 10;
static const u_int CIAddrOffset = 12;
static const u_int YIAddrOffset = 16;
static const u_int SIAddrOffset = 20;
static const u_int GIAddrOffset = 24;
static const u_int CHAddrOffset = 28;
static const u_int SnameOffset = 44;
static const u_int SnameLen = 64;
static const u_int FileOffset = 108; /* SnameOffset + SnameLen */
static const u_int FileLen = 128;
static const u_int CookieOffset = 236; /* FileOffset + FileLen */
static const u_int OptionsOffset = 240; /* CookieOffset+sizeof(OptionsCookie) */
static const u_int MinDHCPLen = 240; /* OptionsOffset */

/* DHCP operations */
static const u_char OpRequest = 1;
static const u_char OpReply = 2;

/* the magic cookie that defines the start of options */
static const u_char OptionsCookie[4] = { 99, 130, 83, 99 };

/* Special options have symbolic names rather than string names */
static const u_int OptionPad = 0;
static const u_int OptionOverload = 52;
static const u_int OptionMessageType = 53;
static const u_int OptionParameterRequest = 55;
static const u_int OptionMaxDHCPMessage = 57;
static const u_int OptionEnd = 255;

/* function prototype for the option parsing function.  Will be called with
 * option's code (code), as well a pointer to the option's data and a value
 * representing the maximum length of the option.
 */
typedef void (*ParserFunc)(u_int code, const u_char *option, u_int len);

struct OptionParser
{
  ParserFunc func;
  const u_char *sampleData;
  u_char sampleLen;
};

/* structure that represents a DHCP option parser */
struct OptionEntry
{
  const char *name;
  struct OptionParser parser;
  Bool parse;
};

/* space to store all possible options.  Note that OptionPad(0) and
 * OptionEnd(255) are treated as special cases, so they won't have parsers
 * here.  Leaving space just to be extra safe.
 */
static struct OptionEntry gOptionParsers[MAX_OPTION + 1];

/* macro to define a new option parser, with name optionName, value
 * optionValue, parser optionParser, and whether to parse by default
 * (optionParse, a boolean).  optionValue is assumed to be > 0 and < MAX_OPTION.
 * See the ParserFunc typedef for help defining that function.
 * Every option defined in this way must expect a length--this isn't
 * appropriate for the lengthless options OptionPad or OptionEnd, ergo their
 * exclusion.
 */
#define DEFINE_OPTION(optionName, optionValue, optionParser, optionParse) \
 assert((optionValue) > 0 && (optionValue) < MAX_OPTION); \
 if ((optionValue) > 0 && (optionValue) < MAX_OPTION) \
 { \
   gOptionParsers[(optionValue)].name = #optionName; \
   gOptionParsers[(optionValue)].parser = (optionParser); \
   gOptionParsers[(optionValue)].parse = (optionParse); \
 }

/* meanings of the overload option */
static const u_int OverloadNone = 0x0; /* not specified, useful default value */
/* The following two aren't specified as flags, but conveniently, if both
 * fields are overloaded, the value is 3, so I'll use them as flags.
 */
static const u_int OverloadFile = 0x1;
static const u_int OverloadSname = 0x2;

/* meanings of the message type option */
static const u_char MessageTypeDiscover = 1;
static const u_char MessageTypeOffer = 2;
static const u_char MessageTypeRequest = 3;
static const u_char MessageTypeDecline = 4;
static const u_char MessageTypeACK = 5;
static const u_char MessageTypeNAK = 6;
static const u_char MessageTypeRelease = 7;
static const u_char MessageTypeInform = 8;

/* port numbers */
static const u_short PortServer = 67;
static const u_short PortClient = 68;

/* parser definitions */
static void parseIPOption(u_int code, const u_char *option, u_int len);
static void parseIPPair(u_int code, const u_char *option, u_int len);
static void parseStringOption(u_int code, const u_char *option, u_int len);
static void parseBoolOption(u_int code, const u_char *option, u_int len);
static void parseMessageType(u_int code, const u_char *option, u_int len);
static void parseOctetOption(u_int code, const u_char *option, u_int len);
static void parseShortOption(u_int code, const u_char *option, u_int len);
static void parseIntOption(u_int code, const u_char *option, u_int len);
static void parseOverloadOption(u_int code, const u_char *option, u_int len);
static void parseDirectoryAgent(u_int code, const u_char *option, u_int len);
static void parseServiceScope(u_int code, const u_char *option, u_int len);

struct OptionParser nullParser = { NULL, NULL, 0 };
struct OptionParser overloadParser = { parseOverloadOption, NULL, 0 };
struct OptionParser messageTypeParser = { parseMessageType, NULL, 0 };
const u_char ipData[] = { 192, 168, 1, 1 };
struct OptionParser ipParser = { parseIPOption, ipData, sizeof(ipData) };
const u_char ipPairData[] = { 192, 168, 1, 1, 255, 255, 255, 0 };
struct OptionParser ipPairParser = { parseIPPair, ipPairData,
 sizeof(ipPairData) };
/* TODO:  this example string doesn't make sense in every instance where string
 * parsing is used.  Replace with more specific examples, e.g a URL, a FQDN,
 * etc.
 */
const u_char stringData[] = "example string";
struct OptionParser stringParser = { parseStringOption, stringData,
 sizeof(stringData) };
const u_char boolData[] = { 1 };
struct OptionParser boolParser = { parseBoolOption, boolData,
 sizeof(boolData) };
const u_char octetData[] = { 7 };
struct OptionParser octetParser = { parseOctetOption, octetData,
 sizeof(octetData) };
const u_char shortData[] = { 3, 0x88 };
struct OptionParser shortParser = { parseShortOption, shortData,
 sizeof(shortData) };
const u_char intData[] = { 2, 70, 80, 90 };
struct OptionParser intParser = { parseIntOption, intData, sizeof(intData) };
const u_char directoryAgentData[] = { 1, 10, 0, 0, 1 };
struct OptionParser directoryAgentParser = { parseDirectoryAgent,
 directoryAgentData, sizeof(directoryAgentData) };
const u_char serviceScopeData[] = { 0, 'e', 'x', 'a', 'm', 'p', 'l', 'e', 'S',
 'c', 'o', 'p', 'e' };
struct OptionParser serviceScopeParser = { parseServiceScope, serviceScopeData,
 sizeof(serviceScopeData) };

/* Initializes global table gOptionParsers.  Call once per application.
 * All definitions can be found at:
   www.iana.org/assignments/bootp-dhcp-extensions/
 */
static void defineOptions(void)
{
  memset(gOptionParsers, 0, sizeof(gOptionParsers));
  DEFINE_OPTION(SubnetMask, 1, ipParser, False);
  DEFINE_OPTION(TimeOffset, 2, intParser, False);
  DEFINE_OPTION(Router, 3, ipParser, False);
  DEFINE_OPTION(TimeServer, 4, ipParser, False);
  DEFINE_OPTION(NameServer, 5, ipParser, False);
  DEFINE_OPTION(DNSServer, 6, ipParser, False);
  DEFINE_OPTION(LogServer, 7, ipParser, False);
  DEFINE_OPTION(CookieServer, 8, ipParser, False);
  DEFINE_OPTION(LPRServer, 9, ipParser, False);
  DEFINE_OPTION(ImpressServer, 10, ipParser, False);
  DEFINE_OPTION(ResourceLocationServer, 11, ipParser, False);
  DEFINE_OPTION(HostName, 12, stringParser, False);
  DEFINE_OPTION(BootFileSize, 13, shortParser, False);
  DEFINE_OPTION(MeritDumpFile, 14, stringParser, False);
  DEFINE_OPTION(DomainName, 15, stringParser, False);
  DEFINE_OPTION(SwapServer, 16, ipParser, False);
  DEFINE_OPTION(RootPath, 17, stringParser, False);
  DEFINE_OPTION(ExtensionsPath, 18, stringParser, False);
  DEFINE_OPTION(IPForwarding, 19, boolParser, False);
  DEFINE_OPTION(SourceRouting, 20, boolParser, False);
  DEFINE_OPTION(PolicyFilter, 21, ipPairParser, False);
  DEFINE_OPTION(MaxDatagramReassembly, 22, shortParser, False);
  DEFINE_OPTION(DefaultTTL, 23, octetParser, False);
  DEFINE_OPTION(PathMTUTimeout, 24, intParser, False);
  DEFINE_OPTION(PathMTUPlateau, 25, nullParser, False); /* short list */
  DEFINE_OPTION(InterfaceMTU, 26, shortParser, False);
  DEFINE_OPTION(AllSubnetsLocal, 27, boolParser, False);
  DEFINE_OPTION(BroadcastAddress, 28, ipParser, False);
  DEFINE_OPTION(PerformMaskDiscovery, 29, boolParser, False);
  DEFINE_OPTION(MaskSupplier, 30, boolParser, False);
  DEFINE_OPTION(PerformRouterDiscovery, 31, boolParser, False);
  DEFINE_OPTION(RouterSolicitationAddr, 32, ipParser, False);
  DEFINE_OPTION(StaticRoute, 33, ipPairParser, False);
  DEFINE_OPTION(TrailerEncap, 34, boolParser, False);
  DEFINE_OPTION(ARPTimeout, 35, intParser, False);
  DEFINE_OPTION(EthernetEncap, 36, boolParser, False);
  DEFINE_OPTION(TCPTTL, 37, octetParser, False);
  DEFINE_OPTION(TCPKeepalive, 38, intParser, False);
  DEFINE_OPTION(TCPKeepaliveGarbage, 39, boolParser, False);
  DEFINE_OPTION(NISDomain, 40, stringParser, False);
  DEFINE_OPTION(NISServer, 41, ipParser, False);
  DEFINE_OPTION(NTPServer, 42, ipParser, False);
  DEFINE_OPTION(VendorSpecific, 43, nullParser, False);
  DEFINE_OPTION(NetBIOSNameServer, 44, ipParser, False);
  DEFINE_OPTION(NetBIOSDDServer, 45, ipParser, False);
  DEFINE_OPTION(NetBIOSNodeType, 46, octetParser, False);
  DEFINE_OPTION(NetBIOSScope, 47, stringParser, False);
  DEFINE_OPTION(XFontServer, 48, ipParser, False);
  DEFINE_OPTION(XDisplayManager, 49, ipParser, False);
  DEFINE_OPTION(RequestedIP, 50, ipParser, False);
  DEFINE_OPTION(IPLeaseTime, 51, intParser, False);
  DEFINE_OPTION(Overload, 52, overloadParser, True);
  DEFINE_OPTION(MessageType, 53, messageTypeParser, True);
  DEFINE_OPTION(ServerIdentifier, 54, ipParser, False);
  DEFINE_OPTION(ParameterRequest, 55, nullParser, False); /* octet list */
  DEFINE_OPTION(Message, 56, stringParser, False);
  DEFINE_OPTION(MaxDHCPMessage, 57, shortParser, False);
  DEFINE_OPTION(RenewalTime, 58, intParser, False);
  DEFINE_OPTION(RebindingTime, 59, intParser, False);
  DEFINE_OPTION(VendorClass, 60, nullParser, False); /* octet list */
  DEFINE_OPTION(ClientIdentifier, 61, nullParser, False); /* octet list */
  DEFINE_OPTION(NetwareIPDomain, 62, stringParser, False);
  DEFINE_OPTION(NetwareIPOption, 63, nullParser, False); /* in RFC2242 */
  DEFINE_OPTION(NISPlusDomain, 64, stringParser, False);
  DEFINE_OPTION(NISPlusServer, 65, ipParser, False);
  DEFINE_OPTION(TFTPServer, 66, stringParser, False);
  DEFINE_OPTION(Bootfile, 67, stringParser, False);
  DEFINE_OPTION(MobileIPHomeAgent, 68, ipParser, False);
  DEFINE_OPTION(SMTPServer, 69, ipParser, False);
  DEFINE_OPTION(POPServer, 70, ipParser, False);
  DEFINE_OPTION(NNTPServer, 71, ipParser, False);
  DEFINE_OPTION(DefaultWWWServer, 72, ipParser, False);
  DEFINE_OPTION(DefaultFingerServer, 73, ipParser, False);
  DEFINE_OPTION(DefaultIRCServer, 74, ipParser, False);
  DEFINE_OPTION(StreetTalkServer, 75, ipParser, False);
  DEFINE_OPTION(StreetTalkDAServer, 76, ipParser, False);
  DEFINE_OPTION(UserClass, 77, nullParser, False); /* in RFC3004 */
  DEFINE_OPTION(DirectoryAgent, 78, directoryAgentParser, False);
  DEFINE_OPTION(ServiceScope, 79, serviceScopeParser, False);
  DEFINE_OPTION(NamingAuthority, 80, stringParser, False);
  DEFINE_OPTION(ClientFQDN, 81, stringParser, False);
  DEFINE_OPTION(RelayAgent, 82, nullParser, False); /* in RFC3046 */
  DEFINE_OPTION(AgentRemoteID, 83, nullParser, False);
  DEFINE_OPTION(AgentSubnetMask, 84, nullParser, False);
  DEFINE_OPTION(NDSServers, 85, ipParser, False);
  DEFINE_OPTION(NDSTreeName, 86, stringParser, False); /* UTF-8 string */
  DEFINE_OPTION(NDSContext, 87, stringParser, False); /* UTF-8 string */
  DEFINE_OPTION(IEEETimezone, 88, stringParser, False);
  DEFINE_OPTION(FQDN, 89, stringParser, False);
  DEFINE_OPTION(Authentication, 90, nullParser, False); /* in RFC3118 */
  DEFINE_OPTION(VinesServer, 91, ipParser, False);
  DEFINE_OPTION(ServerSelection, 92, shortParser, False);
  DEFINE_OPTION(ClientSystem, 93, nullParser, False);
  DEFINE_OPTION(ClientNDI, 94, nullParser, False);
  DEFINE_OPTION(LDAP, 95, stringParser, False);
  DEFINE_OPTION(IPv6Transition, 96, ipParser, False);
  DEFINE_OPTION(UUID, 97, nullParser, False);
  DEFINE_OPTION(UserAuth, 98, stringParser, False); /* space-separated URLs */
  DEFINE_OPTION(PrinterName, 100, stringParser, False);
  DEFINE_OPTION(MDHCP, 101, nullParser, False);
  DEFINE_OPTION(SwapPath, 108, nullParser, False); /* string? */
  DEFINE_OPTION(IPXCompatibility, 110, nullParser, False);
  DEFINE_OPTION(NetinfoAddress, 112, ipParser, False);
  DEFINE_OPTION(NetinfoTag, 113, nullParser, False); /* still don't know.. */
  DEFINE_OPTION(URL, 114, stringParser, False); /* default web URL */
  DEFINE_OPTION(Failover, 115, nullParser, False);
  DEFINE_OPTION(AutoConfig, 116, boolParser, False);
  DEFINE_OPTION(NameServiceSearch, 117, nullParser, False); /* in RFC2937 */
  DEFINE_OPTION(SubnetSelectionOption, 118, ipParser, False);
  DEFINE_OPTION(DomainSearch, 119, nullParser, False); /* in RFC3396 */
  DEFINE_OPTION(SIPServers, 120, nullParser, False); /* in RFC3361 */
  DEFINE_OPTION(ClasslessStaticRoute, 121, nullParser, False); /* in RFC3442 */
  DEFINE_OPTION(WPAD, 252, stringParser, False);
}

/* Helper function the returns the name of the DHCP code given.
 * If code is an invalid or unknown code, prints the string "Unknown"
 */
static const char *printCodeName(u_int code)
{
  const char *name = NULL;

  if (code < MAX_OPTION)
    name = gOptionParsers[code].name;
  if (name == NULL)
    name = "Unknown";
  return name;
}

/* Looks at the option data pointed to by option, and prints out any IP
 * values as dotted decimals to stdout.  Assumes option is not NULL, len is
 * divisible by 4, and every 4-byte value is an IP address in network order.
 */
static void parseIPOption(u_int code, const u_char *option, u_int len)
{
  assert(option != NULL);
  if (option != NULL && len > 0)
  {
    if (!gQuiet)
      printf("%s ", printCodeName(code));
    while (len >= 4)
    {
      printf("%d.%d.%d.%d", option[0], option[1], option[2], option[3]);
      len -= 4;
      option += 4;
      if (len > 0)
        putchar(' ');
    }
    putchar('\n');
  }
}

/* Looks at the option data pointed to by option, and prints it out as a
 * list of (IP/IP) dotted decimal IP pairs to stdout.  Assumes option is not
 * NULL, len is divisible by 8, and every 4 bytes represent an IP address in
 * network order.
 */
static void parseIPPair(u_int code, const u_char *option, u_int len)
{
  assert(option != NULL);
  if (option != NULL && len > 0)
  {
    if (!gQuiet)
      printf("%s ", printCodeName(code));
    while (len >= 8)
    {
      printf("(%d.%d.%d.%d/%d.%d.%d.%d)",
      option[0], option[1], option[2], option[3],
      option[4], option[5], option[6], option[7]);
      len -= 8;
      option += 8;
      if (len > 0)
        putchar(' ');
    }
    putchar('\n');
  }
}

/* Looks at the option data pointed to by option, and prints out the value
 * as an ASCII string.  Assumes option is not NULL.
 */
static void parseStringOption(u_int code, const u_char *option, u_int len)
{
  assert(option != NULL);
  if (option != NULL && len > 0)
  {
    if (!gQuiet)
      printf("%s ", printCodeName(code));
    while (len > 0)
    {
      putchar(option[0]);
      option++;
      len--;
    }
    putchar('\n');
  }
}

/* Looks at the option data pointed to by option, and prints it out as a
 * boolean to stdout.  Assumes option is not NULL, and len is 1.
 */
static void parseBoolOption(u_int code, const u_char *option, u_int len)
{
  assert(option != NULL);
  if (option != NULL && len > 0)
  {
    if (!gQuiet)
      printf("%s ", printCodeName(code));
    printf("%s\n", *option == 1 ? "true" : "false");
  }
}

/* Looks at the option pointed to by option, and interprets it as a single
 * octet meaning message type.  Saves the type in gMessageType.
 * Assumes option is not NULL, and len is 1.  If in verbose mode,
 * prints out the message type as well.
 */
static void parseMessageType(u_int code, const u_char *option, u_int len)
{
  assert(option != NULL);
  if (option != NULL && len > 0)
  {
    gMessageType = *option;
    if (gVerbose)
      printf("%s %d\n", printCodeName(code), *option);
  }
}

/* Looks at the option data pointed to by option, and prints it out as a single
 * octet to stdout.  Assumes option is not NULL, and len is 1.
 */
static void parseOctetOption(u_int code, const u_char *option, u_int len)
{
  assert(option != NULL);
  if (option != NULL && len > 0)
  {
    if (!gQuiet)
      printf("%s ", printCodeName(code));
    printf("%d\n", *option);
  }
}

/* Interprets the option data at *option as the overload option, and saves it
 * in gOverload.  Assumes option is not NULL, and len is 1.
 * If in verbose mode, prints out the overload value as well.
 */
static void parseOverloadOption(u_int code, const u_char *option, u_int len)
{
  assert(option != NULL);
  if (option != NULL && len > 0)
  {
    if (gVerbose)
      printf("%s %d\n", printCodeName(code), *option);
    gOverload = *option;
  }
}

/* Looks at the option data pointed to by option, and prints it out as a short
 * integer value to stdout.  Assumes option is not NULL, and len is 2.
 */
static void parseShortOption(u_int code, const u_char *option, u_int len)
{
  assert(option != NULL);
  if (option != NULL && len >= 2)
  {
    u_short val;

    memcpy(&val, option, sizeof(val));
    if (!gQuiet)
      printf("%s ", printCodeName(code));
    printf("%d\n", ntohs(val));
  }
}

/* Looks at the option data pointed to by option, and prints it out as an
 * integer value to stdout.  Assumes option is not NULL, and len is 4.
 */
static void parseIntOption(u_int code, const u_char *option, u_int len)
{
  assert(option != NULL);
  if (option != NULL && len >= 4)
  {
    u_long val;

    memcpy(&val, option, sizeof(val));
    if (!gQuiet)
      printf("%s ", printCodeName(code));
    printf("%ld\n", (long)ntohl(val));
  }
}

/* From RFC2610, the SLP directory agent.  Assumes option is not NULL, and len
 * is at least 1 and len - 1 is divisible by 4.  The first byte is printed as
 * the mandatory flag, and subsequent 4-byte entities are printed as dotted-
 * decimal IPs.
 */
static void parseDirectoryAgent(u_int code, const u_char *option, u_int len)
{
  assert(option != NULL);
  if (option != NULL && len >= 1)
  {
    if (!gQuiet)
      printf("%s ", printCodeName(code));
    printf("Mandatory=%d ", option[0]);
    option++;
    len--;
    while (len >= 4)
    {
      printf("%d.%d.%d.%d", option[0], option[1], option[2], option[3]);
      len -= 4;
      option += 4;
      if (len > 0)
        putchar(' ');
    }
    putchar('\n');
  }
}

/* From RFC2610, the SLP service scope.  Assumes option is not NULL, and len
 * is at least 1.  The first byte is printed as the mandatory flag, and
 * the remaining bytes are printed to stdout as an ASCII string.  (They're
 * UTF-8 encoded, the caller can decode them if needed.)
 */
static void parseServiceScope(u_int code, const u_char *option, u_int len)
{
  assert(option != NULL);
  if (option != NULL && len >= 1)
  {
    if (!gQuiet)
      printf("%s ", printCodeName(code));
    printf("Mandatory=%d ", option[0]);
    option++;
    len--;
    while (len > 0)
    {
      putchar(option[0]);
      len--;
      option++;
    }
    putchar('\n');
  }
}

/* Parses the options starting at buf, assumed to non-NULL.  Continues parsing
 * until len bytes have been read, or the end option has been found.
 * For each option, if there is an entry in gOptionParsers for it, and if that
 * entry has parse set to True, and if the parser value of that entry is not
 * NULL, calls the parser with the current option's data.
 * If gVerbose is True, prints out the name of every option it encounters,
 * but does not call the parser unless the above checks would cause it to be
 * called.
 */
static void parseOptions(const u_char *buf, u_int len)
{
  u_int offset;
  Bool end = False;

  assert(buf != NULL);
  if (buf == NULL)
    return;

  for (offset = 0; offset < len && !end; )
  {
    if (buf[offset] == OptionPad)
    {
      if (gVerbose)
        printf("Pad\n");
      offset++;
    }
    else if (buf[offset] == OptionEnd)
    {
      if (gVerbose)
        printf("End\n");
      end = True;
    }
    else if (buf[offset] < MAX_OPTION)
    {
      struct OptionEntry *entry = &gOptionParsers[buf[offset]];

      if (entry->parse && entry->parser.func != NULL)
      {
        u_char optionLen;

        if (offset + 1 < len)
          optionLen = min(len - offset - 2, *(buf + offset + 1));
        else
          optionLen = 0;
        entry->parser.func(buf[offset], buf + offset + 2, optionLen);
      }
      else if (gVerbose)
        printf("(Ignoring option %s)\n", printCodeName(buf[offset]));
      offset++;
      if (offset < len)
        offset += buf[offset] + 1;
    }
  }
}

/* Parses the DHCP packet for requested options and prints them out.
 * Assumes pkt is not NULL and that length is at least MinDHCPLen.
 * Returns whether the parsed packet is an ACK that matches our XID (and
 * therefore is the response we're looking for).
 */
static Bool parseDHCP(const u_char *pkt, u_int len)
{
  Bool done = False;
  u_int xid;

  assert(pkt != NULL);
  if (pkt == NULL)
    return False;
  assert(len >= MinDHCPLen);
  if (len < MinDHCPLen)
    return False;

  if (gVerbose)
    printf("Parsing message with length %d\n", len);
  memcpy(&xid, pkt + XidOffset, sizeof(xid));
  /* leave in network order, as we stored gXID in network order */
  if (xid == gXID)
  {
    parseOptions(pkt + OptionsOffset, len - OptionsOffset);
    if (gMessageType == MessageTypeACK)
    {
      if (gOverload & OverloadFile)
        parseOptions(pkt + FileOffset, FileLen);
      if (gOverload & OverloadSname)
        parseOptions(pkt + SnameOffset, SnameLen);
      done = True;
    }
    else
    {
      if (gVerbose)
        printf("Message type isn't an ACK, ignoring\n");
    }
  }
  else
  {
    if (gVerbose)
      printf("xid %08x doesn't match expected (%08x), ignoring\n", xid, gXID);
  }
  return done;
}

/* Prints the usage of this program; arg is assumed to be a NULL-terminated
 * string representing the program name (argv[0] will do just fine).
 */
static void printUsage(const char *arg)
{
  assert(arg != NULL);
  if (arg == NULL)
    return;
  printf(
   "Usage: %s [-d dest1] [-d dest2...] [-q | -v] [-h [options]] [options]\n",
   arg);
  printf(
   "  -d: destination to send to, in the form [server]@<interface>, where:\n");
  printf(
   "      - server is IP address of the DHCP server\n");
  printf(
   "      - interface is the interface with which to send, e.g. eth0\n");
  printf(
   "      If the server is specified, sends a unicast packet to server\n");
  printf(
   "      using the specified interface.  If not specified, sends a\n");
  printf(
   "      broadcast to the interface's broadcast address.  If no\n");
  printf(
   "      destinations are specified, will broadcast to every interface\n");
  printf(
   "      that supports it, and return options from the first server that\n");
  printf(
   "      responds.\n\n");
  printf(
   "      Example 1:  '-d @eth0' sends a broadcast to eth0\n");
  printf(
   "      Example 2:  '-d 192.168.1.1@eth0' sends a unicast to 192.168.1.1\n");
  printf(
   "                  on interface eth0\n\n");
  printf(
   "  -q: quiet mode (don't print option codes/names, just values)\n");
  printf(
   "  -v: verbose mode\n\n");
  printf(
   "  -h:             print usage (this output)\n");
  printf(
   "  -h options:     prints all available options\n");
  printf(
   "  -h op1 op2 ...: prints example output from each requested option\n\n");
  printf(
   "  op1, op2, etc. are DHCP options you wish to request.  These may\n");
  printf(
   "  either be numeric, e.g. 15 (for the DomainName option), or an option\n");
  printf(
   "  name.\n");
  printf(
   "  Named are matched without regard to case.\n");
  printf(
   "\nIf you see an error like, \"bind: Permission denied\", you need to\n");
  printf(
   "run as root, or make the program suid root.  bind(2) needs to be root\n");
  printf(
   "to bind to the DHCP client port.\n");
}

/* UNIX-specific function to "drop permissions"--literally, change the
 * effective UID to the real UID.
 */
static void dropPermissions(void)
{
#ifndef WIN32
  seteuid(getuid());
#endif
}

/* UNIX-specific function to "raise permissions"--literally, change the
 * effective user ID to 0 (root).
 */
static void raisePermissions(void)
{
#ifndef WIN32
  seteuid(0);
#endif
}

/* Adds an option to the DHCP packet at address buf.  bufLen indicates how
 * many bytes may be stored at buf.  currentLen is the current number of
 * bytes stored at buf.  option is the option code, optionLen is the number
 * of bytes from optionData to copy.
 * Assumes buf is not NULL, bufLen is at the very least MinDHCPLen bytes.
 * If option is 0 or 255, only the option code is stored, and currentLen + 1
 * must be less than bufLen.
 * If option is any other value, the option code and optionLen are stored,
 * and currentLen + optionLen + 2 must be less than bufLen.
 * If optionLen is greater than 0, assumes that optionData is not NULL.
 * Returns 0 on success (after which the caller should take care to update
 * currentLen to currentLen + optionLen + 2), -1 on failure.
 */
static int addOption(u_char *buf, u_int bufLen, u_int currentLen,
 u_int option, u_int optionLen, u_char *optionData)
{
  int ret;

  assert(buf != NULL);
  if (buf == NULL)
    return -1;
  assert(bufLen > MinDHCPLen);
  if (bufLen <= MinDHCPLen)
    return -1;
  assert(currentLen + 1 <= bufLen);
  if (currentLen + 1 > bufLen)
    return -1;

  buf[currentLen] = option;
  if (option != 0 && option != 255)
  {
    assert(currentLen + optionLen + 2 <= bufLen);
    if (currentLen + optionLen + 2 > bufLen)
      ret = -1;
    else
    {
      buf[currentLen + 1] = optionLen;
      if (optionLen > 0)
      {
        assert(optionData != NULL);
        if (optionData == NULL)
          ret = -1;
        else
        {
          memcpy(buf + currentLen + 2, optionData, optionLen);
          ret = 0;
        }
      }
      else
        ret = 0;
    }
  }
  else
    ret = 0;
  return ret;
}

/* Gets a value for an XID from the current time.  Not particularly random, but
 * there's no cryptographic requirement for this thing nor is there a session
 * to hijack.
 */
static u_int getXID(void)
{
  struct timeval tv;

  gettimeofday(&tv, NULL);
  return tv.tv_usec;
}

/* Creates the DHCPINFORM request for us at buffer buf with length *bufLen.
 * Uses localaddr as the local IP address, and mac as the local hardware
 * address.  maxResponseLen is the maximum size of DHCP response you (the
 * caller) are prepared to receive, in host order.  Uses gNumRequestedOptions
 * and gRequestedOptions to fill in the ParameterRequest option in the request.
 * Assumes buf and bufLen are not NULL, and *bufLen is at the very least
 * MinDHCPLen long.  Assumes localaddr is in network order.
 * On success, *bufLen will be the new length of the buffer.
 * Returns 0 on success and -1 on failure.
 */
static int createRequest(u_char *buf, u_short *bufLen, u_int localaddr,
 const u_char mac[ETH_ALEN], u_int maxResponseLen)
{
  int ret;
  u_short currentLen = MinDHCPLen;
  u_char op;

  assert(buf != NULL);
  if (buf == NULL)
    return -1;
  assert(bufLen != NULL);
  if (bufLen == NULL)
    return -1;
  assert(*bufLen > MinDHCPLen);
  if (*bufLen <= MinDHCPLen)
    return -1;
  assert(mac != NULL);
  if (mac == NULL)
    return -1;

  buf[OpOffset]    = OpRequest;
  buf[HTypeOffset] = ARPHRD_ETHER;
  buf[HLenOffset]  = ETH_ALEN;
  buf[HOpsOffset]  = 0;
  gXID = htonl(getXID());
  memcpy(buf + XidOffset, &gXID, sizeof(gXID));
  memset(buf + SecsOffset, 0, sizeof(u_short)); /* secs since start */
  memset(buf + FlagsOffset, 0, sizeof(u_short)); /* flags: we accept unicast */
  memcpy(buf + CIAddrOffset, &localaddr, sizeof(localaddr));
  memset(buf + YIAddrOffset, 0, sizeof(u_int)); /* yiaddr */
  memset(buf + SIAddrOffset, 0, sizeof(u_int)); /* siaddr */
  memset(buf + GIAddrOffset, 0, sizeof(u_int)); /* giaddr */
  memcpy(buf + CHAddrOffset, mac, ETH_ALEN);
  memset(buf + SnameOffset, 0, SnameLen);
  memset(buf + FileOffset, 0, FileLen);
  memcpy(buf + CookieOffset, OptionsCookie, sizeof(OptionsCookie));

  op = MessageTypeInform;
  ret = addOption(buf, *bufLen, currentLen, OptionMessageType, sizeof(op), &op);
  if (ret == 0)
  {
    u_short maxLength = htons((u_short)maxResponseLen);

    currentLen += sizeof(op) + 2;
    ret = addOption(buf, *bufLen, currentLen, OptionMaxDHCPMessage,
     sizeof(maxLength), (u_char *)&maxLength);
    if (ret == 0)
    {
      currentLen += sizeof(maxLength) + 2;
      ret = addOption(buf, *bufLen, currentLen, OptionParameterRequest,
       gNumRequestedOptions, gRequestedOptions);
      if (ret == 0)
      {
        currentLen += gNumRequestedOptions + 2;
        ret = addOption(buf, *bufLen, currentLen, OptionEnd, 0, NULL);
        if (ret == 0)
        {
          currentLen += 0 + 1;
          *bufLen = currentLen;
        }
      }
    }
  }
  return ret;
}

/* Prepares a socket to send a DHCP request with local IP address localaddr,
 * and ready for broadcast if broadcast is set.
 * Assumes localaddr is in network order.
 * Returns the socket descriptor (greater than 0) if successful, -1 on error.
 */
static int prepareSocket(u_int localaddr, Bool broadcast)
{
  int ret, sock = socket(PF_INET, SOCK_DGRAM, 0);

  if (sock == -1)
  {
    perror("socket");
    ret = sock;
  }
  else
  {
    int opt = 1;

    /* in case a DHCP client is running, reuse the socket */
    ret = setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,
     (SetSockOptValueType *)&opt, sizeof(opt));
    if (ret == -1)
    {
      perror("setsockopt SO_REUSEADDR");
      close(sock);
    }
    if (broadcast)
    {
      ret = setsockopt(sock, SOL_SOCKET, SO_BROADCAST,
       (SetSockOptValueType *)&opt, sizeof(opt));
      if (ret == -1)
      {
        perror("setsockopt SO_BROADCAST");
        close(sock);
      }
    }
    if (ret == 0)
    {
      struct sockaddr_in addr;

      memset(&addr, 0, sizeof(addr));
      addr.sin_family = AF_INET;
      addr.sin_port = htons(PortClient);
      addr.sin_addr.s_addr = localaddr;

      raisePermissions();
      ret = bind(sock, (struct sockaddr *)&addr, sizeof(addr));
      dropPermissions();
      if (ret == -1)
      {
        perror("bind");
        close(sock);
      }
      else
        ret = sock; /* ollkorrekt, return socket to use now */
    }
  }
  return ret;
}

/* This function:
 * - prepares a socket on which to send a DHCPINFORM request (bound to
 *   info's ip) to addr
 * - calls createRequest to create the message, and sends it
 * recvBufLen is the maximum number of bytes you're able to receive.
 * Assumes info and info->name are not NULL, and that recvBufLen <=
 * USHRT_MAX.
 * Returns the socket on which the request was sent on success, nonzero on
 * error.
 */
static int doDHCPInform(const struct InterfaceInfo *info, u_int addr,
 size_t recvBufLen)
{
  int sock;
  struct in_addr destAddr;

  assert(info != NULL);
  if (info == NULL)
    return -1;
  assert(info->name != NULL);
  if (info->name == NULL)
    return -1;
  assert(recvBufLen <= USHRT_MAX);
  if (recvBufLen > USHRT_MAX)
    return -1;

  destAddr.s_addr = addr;

  if (gVerbose)
  {
    struct in_addr srcAddr;

    srcAddr.s_addr = info->ip;
    printf(
     "Using %s (mac %02x:%02x:%02x:%02x:%02x:%02x ip %s), ", info->name,
     info->mac[0], info->mac[1], info->mac[2], info->mac[3], info->mac[4],
     info->mac[5], addrToStr(srcAddr));
    printf("sending to %s\n", addrToStr(destAddr));
  }

  sock = prepareSocket(info->ip, addr == info->bcastAddr);
  if (sock != -1)
  {
    u_char msg[2048];
    u_short msgLen = sizeof(msg);

    if (createRequest(msg, &msgLen, info->ip, info->mac, (u_short)recvBufLen)
     == 0)
    {
      struct sockaddr_in sinAddr;
      int sendtoReturn;

      sinAddr.sin_family = AF_INET;
      sinAddr.sin_port = htons(PortServer);
      sinAddr.sin_addr.s_addr = destAddr.s_addr;
      sendtoReturn = sendto(sock, msg, msgLen, 0, (void *)&sinAddr,
       sizeof(sinAddr));
      if (sendtoReturn < msgLen)
      {
        perror("sendto");
        close(sock);
        sock = -1;
      }
    }
    else
    {
      perror("createRequest");
      close(sock);
      sock = -1;
    }
  }
  return sock;
}

/* Attempts to parse arg as a numeric option code, from 0 to MAX_OPTION.  If
 * valid, it returns the code.  If not, it returns -1.  Assumes arg is not
 * NULL.
 */
static int parseNumericOptionCode(const char *arg)
{
  int ret;

  assert(arg != NULL);
  if (arg != NULL)
  {
    char *endPtr = NULL;
    long option;

    option = strtol(arg, &endPtr, 0);
    if (errno != ERANGE && endPtr && *endPtr == '\0' && option > 0 &&
     option < MAX_OPTION)
    {
      if (option == (long)OptionPad || option == (long)OptionOverload ||
       option == (long)OptionMessageType || option == (long)OptionEnd)
      {
        fprintf(stderr, "Option %s ignored\n", gOptionParsers[option].name);
        ret = -1;
      }
      else if (gOptionParsers[option].parser.func == NULL)
      {
        fprintf(stderr, "Option %s not supported\n",
         gOptionParsers[option].name);
        ret = -1;
      }
      else
        ret = option;
    }
    else
      ret = -1;
  }
  else
    ret = -1;
  return ret;
}

/* Attempts to parse arg as a string option code, whose value makes a string-
 * insensitive match with one of those in gOptionsParsers[*].name.  If the
 * string matches one of these names, returns the code associated with the
 * string.  If not, it returns -1.  Assumes arg is not NULL.
 */
static int parseStringOptionCode(const char *arg)
{
  int ret;

  assert(arg != NULL);
  if (arg != NULL)
  {
    int i;

    ret = -1;
    for (i = 1; i < MAX_OPTION && ret == -1; i++)
    {
      if (gOptionParsers[i].name != NULL)
      {
        /* note to auditors:  I don't use strncasecmp, because I know the
         * first argument to strcasecmp is NULL-terminated--guaranteed by
         * the DEFINE_OPTION macro, since it makes a string literal from its
         * first argument
         */
        if (!strcasecmp(gOptionParsers[i].name, arg))
          ret = i;
      }
    }
  }
  else
    ret = -1;
  return ret;
}

/* Parses the string arg as a DHCP option.  Assumes arg is not NULL.
 * arg may be either the string name of the option as defined in DEFINE_OPTION,
 * or it may the value of the option.
 * If it cannot parse arg as a valid and supported DHCP option, prints an
 * error to stderr.  If it can, has the following side effects:
 * - the option code is added at gRequestedOptions[gNumRequestedOptions]
 * - increments gNumRequestedOptions
 * - gOptionParsers[option code].parse is set to True
 */
static void parseOptionCode(const char *arg)
{
  assert(arg != NULL);
  if (arg != NULL)
  {
    int code;

    if (isdigit((int)arg[0]))
      code = parseNumericOptionCode(arg);
    else
      code = parseStringOptionCode(arg);
    if (code < 0 || code > MAX_OPTION)
      fprintf(stderr, "Invalid option %s\n", arg);
    else if (gNumRequestedOptions >= MAX_OPTION)
      fprintf(stderr, "Option %s:  maximum number of options exceeded\n",
       gOptionParsers[code].name);
    else
    {
      gRequestedOptions[gNumRequestedOptions++] = code;
      gOptionParsers[code].parse = True;
    }
  }
}

/* Checks whether arg is a valid interface in the list gInterfaces.  Returns
 * the associated struct InterfaceInfo * if so, and NULL if not.
 * Assumes name is not NULL.
 */
static struct InterfaceInfo *findInterface(const char *name)
{
  struct InterfaceInfo *ret = NULL;

  assert(name != NULL);
  if (name != NULL)
  {
    struct InterfaceInfo *ptr;

    for (ptr = gInterfaces; ret == NULL && ptr != NULL; ptr = ptr->next)
    {
      if (!strcasecmp(ptr->name, name))
        ret = ptr;
    }
  }
  return ret;
}

/* Parses the argument arg as a destination of the form [server]@<interface>.
 * If it parses incorrectly, prints to stderr.  Otherwise stores the parsed
 * destination information in the related struct Interface in gInterfaces, and
 * marks it in use by setting its sock member to 1.  Also increments
 * gNumDestinations for every valid interface found.
 * Assumes arg is not NULL.
 */
static void parseDestination(char *arg)
{
  assert(arg != NULL);
  if (arg != NULL)
  {
    char *sep = strchr(arg, '@');
    
    if (sep == NULL)
      fprintf(stderr, "-v: bad syntax, expect [server]@<interface>\n");
    else if (gNumDestinations >= FD_SETSIZE)
      fprintf(stderr,
       "%s: maximum number of destinations (%d) exceeded\n", arg, FD_SETSIZE);
    else
    {
      struct InterfaceInfo *iface = findInterface(sep + 1);

      if (iface != NULL)
      {
        if (sep != arg)
        {
          u_int server;

          *sep = '\0';
          server = convertAddr(arg);
          if (server == INADDR_NONE)
            fprintf(stderr, "%s: invalid server IP address\n", arg);
          else
          {
            iface->serverAddr = server;
            iface->sock = 1;
            gNumDestinations++;
          }
        }
        else
        {
          iface->sock = 1;
          gNumDestinations++;
        }
      }
      else
        fprintf(stderr, "%s: invalid interface\n", sep + 1);
    }
  }
}

/* Parses the command line (argc and argv) for arguments.  As side effects,
 * gRequestedOptions, gNumRequestedOptions, and gOptionParsers may be modified.
 * See parseOptionCode and parseDestination for more details.
 */
static void parseCmdLine(int argc, char *argv[])
{
  int i;

  for (i = 1; i < argc && argv[i] != NULL; )
  {
    if (*argv[i] == '-')
    {
      switch (*(argv[i] + 1))
      {
        case 'h':
          gHelp = True;
          if (i + 1 < argc && !strcasecmp("options", argv[i + 1]))
          {
            gPrintOptions = True;
            i += 2;
          }
          else
            i += 1;
          break;
        case 'd':
          if (i + 1 < argc)
            parseDestination(argv[i + 1]);
          else
            fprintf(stderr, "-d: required argument missing\n");
          i += 2;
          break;
        case 'q':
          gQuiet = True;
          if (gVerbose)
            fprintf(stderr, "Can't be both verbose and quiet\n");
          i += 1;
          break;
        case 'v':
          gVerbose = True;
          if (gQuiet)
            fprintf(stderr, "Can't be both verbose and quiet\n");
          i += 1;
          break;
        default:
          fprintf(stderr, "Unknown option -%c\n", *(argv[i] + 1));
          i += 1;
      }
    }
    else
    {
      parseOptionCode(argv[i]);
      i++;
    }
  }
}

/* Prints all supported options to stdout.
 */
static void printAllOptions(void)
{
  u_char i;

  printf("Known options:\n");
  for (i = 1; i < MAX_OPTION; i++)
  {
    if (gOptionParsers[i].name != NULL)
    {
      if (gOptionParsers[i].parser.func != NULL)
        printf("%s\n", printCodeName(i));
      else
        printf("%s: not supported\n", printCodeName(i));
    }
  }
}

/* Prints example data for all the options that have been requested to stdout.
 */
static void printRequestedOptionHelp(void)
{
  u_int i;

  for (i = 0; i < gNumRequestedOptions; i++)
  {
    /* I know this was checked before putting it in, but to be extra defensive
     */
    if (gRequestedOptions[i] > 0 && gRequestedOptions[i] < MAX_OPTION)
    {
      if (gOptionParsers[gRequestedOptions[i]].parser.func != NULL &&
       gOptionParsers[gRequestedOptions[i]].parser.sampleData != NULL &&
       gOptionParsers[gRequestedOptions[i]].parser.sampleLen > 0)
      {
        gOptionParsers[gRequestedOptions[i]].parser.func(gRequestedOptions[i],
         gOptionParsers[gRequestedOptions[i]].parser.sampleData,
         gOptionParsers[gRequestedOptions[i]].parser.sampleLen);
      }
      else
        printf("%s: no example, sorry\n", printCodeName(gRequestedOptions[i]));
    }
  }
}

/* Prints all the options that have been requested to stdout.
 */
static void printRequestedOptions(void)
{
  u_int i;

  for (i = 0; i < gNumRequestedOptions; i++)
    printf("Requesting option %s\n", printCodeName(gRequestedOptions[i]));
}

/* Marks all interfaces in gInterfaces in use by setting their sock member
 * to 1.  Used to simplify mainLoop, which will only have to iterate over
 * gInterfaces and not special-case the no-destinations-specified case.
 */
void makeBroadcastDestinations(void)
{
  struct InterfaceInfo *ptr;

  for (ptr = gInterfaces; ptr != NULL; ptr = ptr->next)
  {
    ptr->sock = 1;
    gNumDestinations++;
  }
}

/* Waits for up to secs seconds for a file descriptor in fds (whose largest set
 * file descriptor is maxFD) to become readable, and reads a packet into buffer
 * buf (with length bufLen) on the first available read packet.  Attempts to
 * parse the packet.
 * Returns when secs seconds have elapsed, or a packet was parsed successfully.
 * Returns 0 on success (packet parsed successfully) or -1 on failure (library
 * call failure or timeout).
 * Assumes fds is not NULL.
 * I know the interface is a little goofy--the caller doesn't need the buffer
 * buf, so why isn't it local?  Answer: because the send request has the
 * sizeof(buf) in it.  Why don't I distinguish timeout from error?  Answer:
 * because I expect error will be relatively uncommon, and the code run on
 * timeout isn't harmful if run on error instead.
 */
int doWait(fd_set *fds, int maxFD, int secs, u_char *buf, u_int bufLen)
{
  struct timeval tv, endTime;
  Bool done = False;
  int ret;

  assert(fds != NULL);
  if (fds == NULL)
    return -1;

  initTimeDelta(&endTime, secs);
  tv.tv_sec = secs;
  tv.tv_usec = 0;
  while (!done)
  {
    ret = select(maxFD + 1, fds, NULL, NULL, &tv);
    if (ret > 0)
    {
      int i;

      for (i = 1; i <= maxFD; i++)
      {
        if (FD_ISSET(i, fds))
        {
          int msgLen = recvfrom(i, buf, bufLen, 0, NULL, 0);

          if (msgLen > 0)
            done = parseDHCP(buf, (u_int)msgLen);
        }
      }
      if (!done)
        adjustTimeDelta(&tv, &endTime);
      ret = 0;
    }
    else
    {
      ret = -1;
      done = True; /* timeout or error, let's go */
    }
  }
  return ret;
}

/* This function is the main driver of the program.  It determines which
 * interfaces on which to send requests, and calls doDHCPInform to send
 * requests on each of them.  It then waits for up to 5 seconds for a response.
 * If one is received it parses it and the program ends, or if 5 seconds expire
 * nothing is done but the program still ends.
 * Ends when either 5 seconds elapses, or parseDHCP returns True on a
 * received message.
 * Depends on the state of certain globals:
 * - gVerbose controls the amount of output, as it does elsewhere
 * - if gNumDestinations is > 0, we send to only the specified interfaces, and
 *   to the servers specified if they are specified.  (See the declaration of
 *   struct Destination for more help.)  Otherwise we send to all interfaces
 *   in gInterfaces.
 * Returns 0 on success, -1 on failure.  Success in this case means we sent
 * the request on each interface and we either timed out or successfully parsed
 * at least one response.
 */
static int mainLoop(void)
{
  fd_set fds;
  int ret, maxFD = 0;
  u_char buf[2048];
  Bool useUnicast = False;
  struct InterfaceInfo *ptr;

  if (gVerbose)
    printRequestedOptions();

  if (gNumDestinations == 0)
    makeBroadcastDestinations();

  ret = 0;
  FD_ZERO(&fds);
  for (ptr = gInterfaces; ptr != NULL; ptr = ptr->next)
  {
    if (ptr->sock != 0)
    {
      if (ptr->serverAddr != INADDR_ANY)
        useUnicast = True;
      ptr->sock = doDHCPInform(ptr, ptr->serverAddr != INADDR_ANY ?
       ptr->serverAddr : ptr->bcastAddr, sizeof(buf));
      if (ptr->sock != -1)
      {
        FD_SET(ptr->sock, &fds);
        maxFD = max(ptr->sock, maxFD);
      }
      else
        ret = -1;
    }
  }

  if (ret == 0 && maxFD > 0)
  {
    ret = doWait(&fds, maxFD, useUnicast ? 1 : 5, buf, sizeof(buf));
    if (ret == -1 && useUnicast)
    {
      ret = 0;
      /* timeout, retry all unicasts as broadcasts */
      for (ptr = gInterfaces; ptr != NULL; ptr = ptr->next)
      {
        if (ptr->serverAddr != INADDR_ANY && ptr->sock > 0)
        {
          close(ptr->sock);
          ptr->serverAddr = INADDR_ANY;
          ptr->sock = doDHCPInform(ptr, ptr->bcastAddr, sizeof(buf));
          if (ptr->sock != -1)
          {
            FD_SET(ptr->sock, &fds);
            maxFD = max(ptr->sock, maxFD);
          }
          else
            ret = -1;
        }
      }
      ret = doWait(&fds, maxFD, 4, buf, sizeof(buf));
    }
  }
  return ret;
}

int main(int argc, char *argv[])
{
  int ret;

  dropPermissions();
  defineOptions();
  WinStartup();
  gInterfaces = findDHCPInterfaces();

  parseCmdLine(argc, argv);
  if (gHelp)
  {
    if (gPrintOptions)
      printAllOptions();
    else if (gNumRequestedOptions > 0)
    {
      printf("Sample output for each option follows:\n");
      printRequestedOptionHelp();
    }
    else
      printUsage(argv[0]);
    ret = 0;
  }
  else
    ret = mainLoop();

  freeInterfaceList(gInterfaces);
  gInterfaces = NULL;
  WinShutdown();
  return ret;
}
