/* Copyright (C) 2003 Juan Lang
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License Version 2 or later
 * (the "GPL").
 *
 * This program is meant to be a simple replacement for nslookup for use by
 * scripts.  If you're writing a C program, gethostbyname or copies of this
 * code are probably for you.
 */
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include "iptypes.h"
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include "iputil.h"
#include "minmax.h"
#include "strcompat.h"
#include "resquery.h"

char *gService = NULL;

struct type_str_match
{
  const char *name;
  int type;
};

/* Only store the ones we match */
static struct type_str_match types[] =
{
  {"A", T_A},
  {"SRV", T_SRV}
};

/* Looks for the name prefix in the domain domain.  If prefix.domain isn't
 * found, looks for a more-general domain to search, excluding TLDs, for
 * prefix (and recurses to do so).
 * If a CNAME is found for the name, prints Name (prefix.founddomain) on one
 * line, followed by Canonical name (canonical name) on the next line.
 * If an A record is found for the name, prints Name (prefix.founddomain) on
 * one line, followed by Addresses (all found IP addresses) on the next line.
 * Returns 0 on success and -1 on failure.
 * Assumes prefix and domain are not NULL.
 */
int simpleFindName(const char *prefix, const char *domain)
{
  int ret;
  char buf[MAX_DOMAIN_NAME];
  struct hostent *hent;

  assert(prefix != NULL);
  if (prefix == NULL)
    return -1;
  assert(domain != NULL);
  if (domain == NULL)
    return -1;

  /* If domain is empty, adding a trailing dot would force this to be a
   * bogus TLD query--only add a dot if domain is not empty.
   */
  if (*domain != '\0')
    snprintf(buf, sizeof(buf), "%s.%s", prefix, domain);
  else
    snprintf(buf, sizeof(buf), "%s", prefix);
  hent = gethostbyname(buf);
  if (hent != NULL)
  {
    printf("Name %s\n", buf);
    /* we know buf is NULL-terminated, but use strncmp to avoid scanner
     * complaints */
    if (strncmp(buf, hent->h_name, sizeof(buf)))
      printf("Canonical name %s\n", hent->h_name);
    else if (hent->h_addr_list != NULL)
    {
      int i;

      if (*hent->h_addr_list != NULL)
        printf("Addresses ");
      for (i = 0; hent->h_addr_list[i] != NULL; i++)
      {
        struct in_addr addr;

        memcpy(&addr.s_addr, hent->h_addr_list[i], sizeof(addr.s_addr));
        printf("%s ", addrToStr(addr));
      }
      if (*hent->h_addr_list != NULL)
        putchar('\n');
    }
    ret = 0;
  }
  else
  {
    char *nextDomain;

    if ((nextDomain = strchr(domain, '.')) != strrchr(domain, '.'))
      ret = simpleFindName(prefix, nextDomain + 1);
    else
      ret = -1;
  }
  return ret;
}

/* Removes the trailing '.' characters from domain.  If type is T_A, calls
 * simpleFindName.  Otherwise calls queryWithType, which it assumes can handle
 * queries of other types.  Assumes prefix and domain are not NULL.
 */
int findName(const char *prefix, char *domain, int type)
{
  int i, ret;

  assert(prefix != NULL);
  if (prefix == NULL)
    return -1;
  assert(domain != NULL);
  if (domain == NULL)
    return -1;

  for (i = strlen(domain) - 1; i >= 0 && domain[i] == '.'; i--)
    domain[i] = '\0';
  if (type == T_A)
    ret = simpleFindName(prefix, domain);
  else
    ret = queryWithType(prefix, domain, type);
  return ret;
}

/* TODO:  describe output a bit in usage
 */
void printUsage(char *prog)
{
  printf("Usage: %s [-t type] <service name>\n", prog);
  printf("  -t: lookup type to use, defaults to A\n");
  printf("  service name: name of prefix to search for, e.g. mail\n");
}

int gLookupType = T_A;

void parseQueryType(const char *arg)
{
  int i, foundType = -1;

  for (i = 0; i < sizeof(types) / sizeof(types[0]) && foundType == -1; i++)
    if (!strcasecmp(arg, types[i].name))
      foundType = i;
  if (foundType != -1)
    gLookupType = types[foundType].type;
}

void parseCmdLine(int argc, char *argv[])
{
  int i;

  for (i = 1; i < argc && argv[i] != NULL; )
  {
    if (*argv[i] == '-')
    {
      switch(*(argv[i] + 1))
      {
        case 't':
          if (i + 1 < argc)
            parseQueryType(argv[i + 1]);
          i += 2;
          break;
        default:
          fprintf(stderr, "Unknown option -%c\n", *(argv[i] + 1));
          i += 1;
      }
    }
    else
    {
      gService = argv[i];
      i += 1;
    }
  }
}

void getMyDomainName(char *domain, u_int domainLen)
{
  char buf[MAX_DOMAIN_NAME];

  assert(domain != NULL);
  if (domain == NULL)
    return;
  assert(domainLen > 0);
  if (domainLen < 0)
    return;

  if (gethostname(buf, sizeof(buf)) == 0)
  {
    struct hostent *hent = gethostbyname(buf);

    if (hent != NULL)
    {
      char *dot = strchr(hent->h_name, '.');

      if (dot != NULL)
      {
        u_int lenToCopy = min(domainLen - 1, strlen(dot + 1));

        memcpy(domain, dot + 1, lenToCopy);
        domain[lenToCopy] = '\0';
      }
    }
    else
      perror("gethostbyname");
  }
  else
    perror("gethostname");
}

int main(int argc, char *argv[])
{
  char domain[MAX_DOMAIN_NAME];
  int ret;

  WinStartup();
  parseCmdLine(argc, argv);
  getMyDomainName(domain, sizeof(domain));
  if (gService == NULL)
  {
    printUsage(argv[0]);
    ret = -1;
  }
  else
    ret = findName(gService, domain, gLookupType);
  WinShutdown();
  return ret;
}
