#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <CoreServices/CoreServices.h>
#include <SystemConfiguration/SystemConfiguration.h>
#include "ifdetect.h"

/* This is the MacOSX implementation of FindDHCPInterfaces et al.  It's
 * implemented using the Service Configuration Framework, and using sample code
 * from developer.apple.com, especially the MoreSCF sample.
 * Being the first time I've used CoreFoundation etc., I don't guarantee this
 * code is free of leaks.  Its saving grace is it's intended to be used in a
 * short-lived program.
 */

static void PrintCFType(const void *value)
{
  if (CFGetTypeID(value) == CFStringGetTypeID())
  {
    CFStringRef str;
    UniChar *printBuf;
    CFRange range;
    int i;

    str = (CFStringRef)value;
    range.location = 0;
    range.length = CFStringGetLength(str);
    printBuf = malloc(sizeof(UniChar) * (range.length));
    CFStringGetCharacters(str, range, printBuf);
    for (i = 0; i < range.length; i++)
      putchar(printBuf[i]);
    free(printBuf);
  }
  else if (CFGetTypeID(value) == CFArrayGetTypeID())
  {
    CFArrayRef array = (CFArrayRef)value;
    int i;

    for (i = 0; i < CFArrayGetCount(value); i++)
      PrintCFType(CFArrayGetValueAtIndex(array, i));
  }
  else if (CFGetTypeID(value) == CFDictionaryGetTypeID())
  {
    CFDictionaryRef dict;
    void **keys, **values;
    int i;

    dict = (CFDictionaryRef)value;
    i = CFDictionaryGetCount(dict);
    keys = malloc(i * sizeof(void *));
    values = malloc(i * sizeof(void *));
    CFDictionaryGetKeysAndValues(dict, (const void **)keys,
     (const void **)values);
    for (i = 0; i < CFDictionaryGetCount(dict); i++)
    {
      printf("Key ");
      PrintCFType(keys[i]);
      printf(": ");
      PrintCFType(values[i]);
      putchar('\n');
    }
    free(keys);
    free(values);
  }
}

/* This function is a callback from GetPartialDHCPAddresses when
 * it calls CFDictionaryApplyFunction.  It extracts the first IP address and
 * netmask and the service ID from each DHCP interface in the network service
 * dictionary and appends it to the result dictionary (which is passed
 * in via the context pointers).
 */
static void AppendDHCPInterfacesToArray(const void *key, const void *value,
 void *context)
{
  CFStringRef configMethod;

  assert( key != NULL );
  assert( CFGetTypeID(key) == CFStringGetTypeID() );
  assert( value != NULL );
  assert( CFGetTypeID(value) == CFDictionaryGetTypeID() );
  assert( context != NULL );
  assert( CFGetTypeID(context) == CFArrayGetTypeID() );

  configMethod = CFDictionaryGetValue(value, kSCPropNetIPv4ConfigMethod);
  if (CFStringCompare(configMethod, CFSTR("DHCP"), kCFCompareCaseInsensitive)
   == kCFCompareEqualTo)
  {
    CFMutableDictionaryRef partial = CFDictionaryCreateMutable(NULL, 0,
     &kCFCopyStringDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks);

    if (partial != nil)
    {
      CFArrayRef keyExploded, intfAddrList, subnetAddrList;

      keyExploded = CFStringCreateArrayBySeparatingStrings(nil, key,
       CFSTR("/"));
      CFDictionarySetValue(partial, CFSTR("ServiceID"),
       CFArrayGetValueAtIndex(keyExploded, 3));
      intfAddrList = CFDictionaryGetValue(value, kSCPropNetIPv4Addresses);
      if (intfAddrList != NULL)
        CFDictionarySetValue(partial, CFSTR("IP"),
         CFArrayGetValueAtIndex(intfAddrList, 0));
      subnetAddrList = CFDictionaryGetValue(value, kSCPropNetIPv4SubnetMasks);
      if (subnetAddrList != NULL)
        CFDictionarySetValue(partial, CFSTR("Netmask"), 
         CFArrayGetValueAtIndex(subnetAddrList, 0));
      CFArrayAppendValue(context, partial);
      CFRelease(keyExploded);
    }
  }
}

/* Returns a CFArray that contains partial configuration information for
 * every DHCP interface on the system (as a CFDictionary) in no
 * particular order.
 * Assumes dynStore is not NULL.
 */
static CFMutableArrayRef GetPartialDHCPInterfaces(SCDynamicStoreRef dynStore)
{
  CFMutableArrayRef result = NULL;

  assert(dynStore != NULL);
  if (dynStore != NULL)
  {
    CFStringRef pattern = SCDynamicStoreKeyCreateNetworkServiceEntity(
     NULL, kSCDynamicStoreDomainSetup, kSCCompAnyRegex, kSCEntNetIPv4);

    if (pattern != NULL)
    {
      CFArrayRef patternList = NULL;

      patternList = CFArrayCreate(NULL, (const void **)&pattern, 1,
       &kCFTypeArrayCallBacks);
      if (patternList != NULL)
      {
        CFDictionaryRef valueDict = SCDynamicStoreCopyMultiple(dynStore,
         NULL, patternList);

        result = CFArrayCreateMutable(NULL, 0, &kCFTypeArrayCallBacks);
        if (result != NULL)
          CFDictionaryApplyFunction(valueDict, AppendDHCPInterfacesToArray,
           result);
        CFRelease(patternList);
      }
      CFRelease(pattern);
    }
  }
  return result;
}

/* Adds the MACAddress value from the CFDictionary value to the CFDictionary
 * context.  Assumes both are not NULL and are CFDictionary types.  Assumes,
 * but doesn't really care, that the two dictionaries are "related", in that
 * they're talking about the same interface.
 */
static void AddMACAddressToInterface(const void *key, const void *value,
 void *context)
{
  CFStringRef macAddr = NULL;

  assert( value != NULL );
  assert( CFGetTypeID(value) == CFDictionaryGetTypeID() );
  assert( context != NULL );
  assert( CFGetTypeID(context) == CFDictionaryGetTypeID() );

  macAddr = CFDictionaryGetValue(value, CFSTR("MACAddress"));
  if (macAddr != NULL)
    CFDictionarySetValue(context, CFSTR("MACAddress"), macAddr);
}

/* For every CFDictionaryRef in the array interfaces, uses the value of key
 * ServiceID to find the associated MAC address of the interface, and adds the
 * key MACAddress to the dictionary.
 */
static void AddMACAddressesToInterfaces(SCDynamicStoreRef dynStore,
 CFMutableArrayRef interfaces)
{
  int i;

  assert(dynStore != NULL);
  if (dynStore == NULL)
    return;
  assert(interfaces != NULL);
  if (interfaces == NULL)
    return;

  for (i = 0; i < CFArrayGetCount(interfaces); i++)
  {
    CFMutableDictionaryRef dict;

    CFArrayGetValues(interfaces, CFRangeMake(i, 1), (const void **)&dict);
    if (dict != NULL && CFGetTypeID(dict) == CFDictionaryGetTypeID())
    {
      CFStringRef serviceID = CFDictionaryGetValue(dict, SCSTR("ServiceID"));

      if (serviceID != NULL)
      {
        CFStringRef pattern = SCDynamicStoreKeyCreateNetworkServiceEntity(
         NULL, kSCDynamicStoreDomainSetup, serviceID, kSCEntNetEthernet);

        if (pattern != NULL)
        {
          CFArrayRef patternList = NULL;

          patternList = CFArrayCreate(NULL, (const void **)&pattern, 1,
           &kCFTypeArrayCallBacks);
          if (patternList != NULL)
          {
            CFDictionaryRef valueDict = NULL;

            valueDict = SCDynamicStoreCopyMultiple(dynStore, NULL, patternList);
            if (valueDict != NULL)
            {
              CFDictionaryApplyFunction(valueDict, AddMACAddressToInterface,
               dict);
            }
            CFRelease(patternList);
          }
          CFRelease(pattern);
        }
      }
    }
  }
}

/* Adds the InterfaceName value from the CFDictionary value to the CFDictionary
 * context.  Assumes both are not NULL and are CFDictionary types.  Assumes,
 * but doesn't really care, that the two dictionaries are "related", in that
 * they're talking about the same interface.
 */
static void AddInterfaceNameToInterface(const void *key, const void *value,
 void *context)
{
  CFStringRef interfaceName = NULL;

  assert( value != NULL );
  assert( CFGetTypeID(value) == CFDictionaryGetTypeID() );
  assert( context != NULL );
  assert( CFGetTypeID(context) == CFDictionaryGetTypeID() );

  interfaceName = CFDictionaryGetValue(value, CFSTR("InterfaceName"));
  if (interfaceName != NULL)
    CFDictionarySetValue(context, CFSTR("InterfaceName"), interfaceName);
}

/* For every CFDictionaryRef in the array interfaces, uses the value of key
 * ServiceID to find the associated interface name, and adds the key Inteface
 * to the dictionary.
 */
static void AddInterfaceNameToInterfaces(SCDynamicStoreRef dynStore,
 CFMutableArrayRef interfaces)
{
  int i;

  assert(dynStore != NULL);
  if (dynStore == NULL)
    return;
  assert(interfaces != NULL);
  if (interfaces == NULL)
    return;

  for (i = 0; i < CFArrayGetCount(interfaces); i++)
  {
    CFMutableDictionaryRef dict;

    CFArrayGetValues(interfaces, CFRangeMake(i, 1), (const void **)&dict);
    if (dict != NULL && CFGetTypeID(dict) == CFDictionaryGetTypeID())
    {
      CFStringRef serviceID = CFDictionaryGetValue(dict, SCSTR("ServiceID"));

      if (serviceID != NULL)
      {
        CFStringRef pattern = SCDynamicStoreKeyCreateNetworkServiceEntity(
         NULL, kSCDynamicStoreDomainState, serviceID, kSCEntNetIPv4);

        if (pattern != NULL)
        {
          CFArrayRef patternList = NULL;

          patternList = CFArrayCreate(NULL, (const void **)&pattern, 1,
           &kCFTypeArrayCallBacks);
          if (patternList != NULL)
          {
            CFDictionaryRef valueDict = NULL;

            valueDict = SCDynamicStoreCopyMultiple(dynStore, NULL, patternList);
            if (valueDict != NULL)
              CFDictionaryApplyFunction(valueDict, AddInterfaceNameToInterface,
               dict);
            CFRelease(patternList);
          }
          CFRelease(pattern);
        }
      }
    }
  }
}

/* Gives you a null-terminated C string from the CStringRef by using an ASCII
 * encoding, or NULL if str is NULL.  You must free() the returned string.
 */
char *StringToASCII(CFStringRef str)
{
  char *cstr;

  if (str != NULL)
  {
    const char *ptr = CFStringGetCStringPtr(str, kCFStringEncodingASCII);

    if (ptr == NULL)
    {
      cstr = (char *)malloc(CFStringGetLength(str) + 1);
      CFStringGetCString(str, cstr, CFStringGetLength(str) + 1,
       kCFStringEncodingASCII);
    }
    else
      cstr = strdup(ptr);
  }
  else
    cstr = NULL;
  return cstr;
}

/* Converts the CFString to an IP address.  Assumes str can be converted to
 * a dotted-decimal IP address string using ASCII encoding.  Returns
 * INADDR_NONE on any error.
 */
unsigned long StringToAddr(CFStringRef str)
{
  unsigned long ret;

  if (str != NULL)
  {
    char *ip = StringToASCII(str);

    if (ip != NULL)
    {
      ret = inet_addr(ip);
      free(ip);
    }
    else
      ret = INADDR_NONE;
  }
  else
    ret = INADDR_NONE;
  return ret;
}

/* Converts strRef to a MAC address in the buffer mac.  Assumes mac is not NULL
 * and strRef can be converted to the form "00:01:02:03:04:05" using ASCII
 * encoding.
 */
void StringToMAC(CFStringRef strRef, u_char mac[ETH_ALEN])
{
  assert(mac != NULL);
  if (strRef != NULL && mac != NULL)
  {
    char *str = StringToASCII(strRef);

    if (str != NULL)
    {
      int byte[ETH_ALEN], i;

      sscanf(str, "%x:%x:%x:%x:%x:%x", &byte[0], &byte[1], &byte[2], &byte[3],
       &byte[4], &byte[5]);
      for (i = 0; i < ETH_ALEN; i++)
        mac[i] = byte[i];
      free(str);
    }
  }
}

struct InterfaceInfo *DictionaryToInterfaceInfo(CFDictionaryRef dict)
{
  struct InterfaceInfo *info;

  assert(dict != NULL);
  if (dict != NULL)
  {
    info = (struct InterfaceInfo *)calloc(1, sizeof(struct InterfaceInfo));
    if (info != NULL)
    {
      u_long mask;

      memset(info, 0, sizeof(struct InterfaceInfo));
      info->name = StringToASCII(CFDictionaryGetValue(dict,
       CFSTR("InterfaceName")));
      info->ip = StringToAddr(CFDictionaryGetValue(dict, CFSTR("IP")));
      mask = StringToAddr(CFDictionaryGetValue(dict, CFSTR("Netmask")));
      info->bcastAddr = (info->ip & mask) | ~mask;
      StringToMAC(CFDictionaryGetValue(dict, CFSTR("MACAddress")), info->mac);
    }
  }
  else
    info = NULL;
  return info;
}

struct InterfaceInfo *findDHCPInterfaces(void)
{
  struct InterfaceInfo *interfaces = NULL;
  SCDynamicStoreRef dynStore;

  dynStore = SCDynamicStoreCreate(NULL, CFSTR("CopyIPAddressListSCF"),
   NULL, NULL);
  if (dynStore != NULL)
  {
    CFMutableArrayRef array = GetPartialDHCPInterfaces(dynStore);
    if (array != NULL)
    {
      int i;

      AddMACAddressesToInterfaces(dynStore, array);
      AddInterfaceNameToInterfaces(dynStore, array);
      for (i = 0; i < CFArrayGetCount(array); i++)
      {
        struct InterfaceInfo *ptr =
         DictionaryToInterfaceInfo(CFArrayGetValueAtIndex(array, i));

        if (ptr != NULL)
        {
          ptr->next = interfaces;
          interfaces = ptr;
        }
      }
      CFRelease(array);
    }
    CFRelease(dynStore);
  }
  return interfaces;
}
