/* Copyright (C) 2003 Juan Lang
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License Version 2 or later
 * (the "GPL").
 *
 * This file is the UNIX implementation file for finding information about
 * interfaces.  See its header file, of course, and dhcpinform.c for where it's
 * used.
 * A note on its implementation: the number of ifdef's is a bit to look at.
 * There are three methods for determining the MAC address of an interface:
 * - a specific IOCTL (Linux)
 * - looking in the ARP cache (at least Solaris)
 * - using the sysctl interface (FreeBSD and MacOSX)
 */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <net/route.h>
#include <sys/ioctl.h>

#if HAVE_SYS_SYSCTL_H
#include <sys/sysctl.h>
#endif

#if HAVE_SYS_SOCKIO_H
#include <sys/sockio.h>
#endif

#if HAVE_NET_IF_DL_H
#include <net/if_dl.h>
#endif

#if HAVE_NET_IF_TYPES_H
#include <net/if_types.h>
#endif

#include "bool.h"
#include "iputil.h"
#include "minmax.h"
#include "ifdetect.h"

static Bool isUpBCastInterface(int sock, const char *name)
{
  Bool ret;
  struct ifreq ifr;

  assert(name != NULL);
  if (name == NULL)
    return 0;

  strncpy(ifr.ifr_name, name, IFNAMSIZ);
  if (ioctl(sock, SIOCGIFFLAGS, &ifr))
  {
    perror("ioctl SIOCGIFFLAGS");
    ret = 0;
  }
  else
    ret = (ifr.ifr_flags & IFF_UP) && (ifr.ifr_flags & IFF_BROADCAST);
  return ret;
}

static u_int getIPAddr(int fd, const char *name)
{
  u_int ret;

  assert(name != NULL);
  if (name != NULL)
  {
    struct ifreq ifr;

    strncpy(ifr.ifr_name, name, IFNAMSIZ);
    if (ioctl(fd, SIOCGIFADDR, &ifr))
    {
      perror("ioctl SIOCGIFADDR");
      ret = INADDR_NONE;
    }
    else
      memcpy(&ret, ifr.ifr_addr.sa_data + 2, sizeof(u_int));
  }
  else
    ret = INADDR_NONE;
  return ret;
}

static u_int getBCastAddr(int fd, const char *name)
{
  u_int ret;

  assert(name != NULL);
  if (name != NULL)
  {
    struct ifreq ifr;

    strncpy(ifr.ifr_name, name, IFNAMSIZ);
    if (ioctl(fd, SIOCGIFBRDADDR, &ifr))
    {
      perror("ioctl SIOCGIFBRDADDR");
      ret = INADDR_NONE;
    }
    else
      memcpy(&ret, ifr.ifr_addr.sa_data + 2, sizeof(u_int));
  }
  else
    ret = INADDR_NONE;
  return ret;
}

#if defined (SIOCGIFHWADDR)
static void getMacAddress(struct InterfaceInfo *info, int fd, const char *name)
{
  struct ifreq ifr;

  memset(&ifr, 0, sizeof(struct ifreq));
  strncpy(ifr.ifr_name, name, IFNAMSIZ);
  if ((ioctl(fd, SIOCGIFHWADDR, &ifr)))
    perror("ioctl SIOCFGIFHWADDR");
  else
    memcpy(info->mac, ifr.ifr_hwaddr.sa_data, ETH_ALEN);
}
#elif defined (SIOCGARP)
static void getMacAddress(struct InterfaceInfo *info, int fd, const char *name)
{
  struct arpreq arp;
  struct sockaddr_in *saddr;
  int ret;

  memset(&arp, 0, sizeof(struct arpreq));
  arp.arp_pa.sa_family = AF_INET;
  saddr = (struct sockaddr_in *)&arp; /* protocol addr is first member of arp */
  saddr->sin_family = AF_INET;
  saddr->sin_addr.s_addr = info->ip;
  if ((ret = ioctl(fd, SIOCGARP, &arp)))
    perror("ioctl SIOCGARP");
  else
    memcpy(info->mac, &arp.arp_ha.sa_data[0], ETH_ALEN);
}
#elif defined (HAVE_SYS_SYSCTL_H) && defined (HAVE_NET_IF_DL_H)
static void getMacAddress(struct InterfaceInfo *info, int fd, const char *name)
{
  struct if_msghdr *ifm;
  struct sockaddr_dl *sdl;
  u_char *p, *buf;
  size_t len;
  int mib[] = { CTL_NET, AF_ROUTE, 0, AF_LINK, NET_RT_IFLIST, 0 };

  if (sysctl(mib, 6, NULL, &len, NULL, 0) < 0)
    return;

  if ((buf = (u_char *)malloc(len)) == NULL)
    return;

  if (sysctl(mib, 6, buf, &len, NULL, 0) < 0)
  {
    free(buf);
    return;
  }
  for (p = buf; p < buf + len; p += ifm->ifm_msglen)
  {
    ifm = (struct if_msghdr *)p;
    sdl = (struct sockaddr_dl *)(ifm + 1);

    if (ifm->ifm_type != RTM_IFINFO || (ifm->ifm_addrs & RTA_IFP) == 0)
      continue;

    if (sdl->sdl_family != AF_LINK || sdl->sdl_nlen == 0 ||
     memcmp(sdl->sdl_data, name, max(sdl->sdl_nlen, strlen(name))) != 0)
      continue;

    memcpy(info->mac, LLADDR(sdl), 6);
    break;
  }
  free(buf);
}
#endif

static struct InterfaceInfo *getInterfaceInfo(const char *name)
{
  struct InterfaceInfo *info;

  assert(name != NULL);
  if (name != NULL)
  {
    int fd = socket (PF_INET, SOCK_DGRAM, 0);

    if (fd != -1)
    {
      info = (struct InterfaceInfo *)calloc(sizeof(struct InterfaceInfo), 1);
      if (info != NULL)
      {
        info->name = strdup(name);
        info->ip = getIPAddr(fd, name);
        info->bcastAddr = getBCastAddr(fd, name);
        getMacAddress(info, fd, name);
      }
      close(fd);
    }
    else
    {
      perror("socket");
      info = NULL;
    }
  }
  else
    info = NULL;
  return info;
}

struct InterfaceInfo *findDHCPInterfaces(void)
{
  const u_int InterfaceIncr = 4;
  int fd = socket (PF_INET, SOCK_DGRAM, 0);
  struct InterfaceInfo *interfaces = NULL;

  if (fd != -1)
  {
    int ret, ndx;
    struct ifconf ifc;
    caddr_t ifPtr;

    ret = 0;
    ndx = 0;
    ifc.ifc_buf = NULL;
    do /* there is no way to know the interface count beforehand,
          so we need to loop again and again upping our max each time
          until returned < max */
    {
      ndx += InterfaceIncr;
      if (ifc.ifc_buf)
        free(ifc.ifc_buf);
      ifc.ifc_len = sizeof(struct ifreq) * ndx;
      ifc.ifc_buf = (char *)malloc((u_int)ifc.ifc_len);
      if ((ret = ioctl(fd, SIOCGIFCONF, &ifc)) < 0)
        perror("ioctl SIOCGIFCONF");
    } while (ret == 0 && ifc.ifc_len == (sizeof(struct ifreq) * ndx));

    ifPtr = ifc.ifc_buf;
    while (ifPtr < ifc.ifc_buf + ifc.ifc_len)
    {
      char ifName[IFNAMSIZ];
      size_t reqLen;

#if HAVE_STRUCT_SOCKADDR_SA_LEN
      reqLen = ((struct ifreq *)ifPtr)->ifr_addr.sa_len + IFNAMSIZ;
#else
      reqLen = sizeof(struct ifreq);
#endif
      memcpy(ifName, ifPtr, IFNAMSIZ);
      if (isUpBCastInterface(fd, ifName))
      {
        struct InterfaceInfo *newIF = getInterfaceInfo(ifName);

        newIF->next = interfaces;
        interfaces = newIF;
      }
      ifPtr += reqLen;
    }
    if (ifc.ifc_buf)
      free(ifc.ifc_buf);
    close(fd);
  }
  else
    perror("socket");
  return interfaces;
}
