/* Copyright (C) 2003 Juan Lang
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License Version 2 or later
 * (the "GPL").
 *
 * This file uses res_querydomain to look up DNS records other than A records.
 * It's mainly used for T_SRV records, but it has some support for other types
 * as well, disabled for now until a use comes up.
 */
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include "iptypes.h"
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include "bool.h"
#include "resquery.h"

typedef struct
{
  u_short  type;
  u_short  class;
  u_int    ttl;
  u_short  length;
} rr_header;

typedef struct
{
  u_short  priority;
  u_short  weight;
  u_short  port;
} srv_header;

/* on input, *location is a pointer to the first byte to parse
 * on return, *location is a pointer after the last byte we consumed
 * print determines whether we want to print the output, or just scan past it
 */
static int decode_name(char** location, char* buffer, int len, Bool print)
{
  u_char*  name_loc = (u_char*)*location;
  u_char*  last_loc = name_loc;
  u_char  name_len;
  u_short  offset;
  int   hit_pointer;

  name_len = *name_loc;

  hit_pointer = False;

  while (name_len != 0)
  {
    if ((name_len & INDIR_MASK) == INDIR_MASK)
    {
      /* If this is the first pointer we've encountered, set 
       * *location to point to after the pointer. */

      if (!hit_pointer)
        *location = (char *) (name_loc + 2);

      hit_pointer = True;

      offset = ((*name_loc) & ~INDIR_MASK) << 8 | *(name_loc + 1);
      name_loc = ((u_char *) buffer) + offset;

      if ((char *) name_loc >= *location)
        return -1;

      if (name_loc == last_loc)
        return -1; /* would yield loop */
       
      last_loc = name_loc;
    }
    else
    {
      int i = 0;

      if (((char *)name_loc) + name_len > buffer + len)
        return -1;

      if (print)
      {
        char name[64];

        for (i = 0; i < name_len; i++)
          name[i] = name_loc[i + 1];
        name[i] = 0;
        printf("%s.", name);
      }

      name_loc += name_len + 1;
    }
    name_len = *name_loc;
  }
    
  /* If the name contains no pointers, then we have to set 
   * *location to point to after the null terminator. */
    
  if (!hit_pointer)
    *location = name_loc + 1;
    
  return 0;
}

static int decode_question(char** location, char* buffer, int len)
{
  if (*location >= buffer + len || *location < buffer)
    return -1;

  printf("Name ");
  if (decode_name(location, buffer, len, True) != 0)
    return -1;
  putchar('\n');

  if (*location + 4 > buffer + len)
    return -1;

  *location += sizeof(u_short);
  *location += sizeof(u_short);

  return 0;
}

static int decode_rr(char** location, char* buffer, int len, int targetType)
{
  rr_header header;
    
  if (*location >= buffer + len || *location < buffer)
    return -1;

  if (decode_name(location, buffer, len, False) != 0)
    return -1;

  if (*location + sizeof(rr_header) >= buffer + len)
    return -1;
    
  memcpy(&header, *location, sizeof(header));

  *location += sizeof(rr_header) - sizeof(header.length);

  if (*location + ntohs(header.length) > buffer + len)
    return -1;

  switch(ntohs(header.type))
  {
    case T_NS:
    case T_CNAME:
    case T_PTR:
    case T_MX:
    {
      char* name = *location;

      if (ntohs(header.type) == targetType)
        printf("  data    : ");
      decode_name(&name, buffer, len, ntohs(header.type) == targetType);
      if (ntohs(header.type) == targetType)
        printf("\n");
      break;
    }
    case T_SRV:
    {
      char* name = *location + sizeof(srv_header);

      if (ntohs(header.type) == targetType)
      {
        srv_header srv;
        
        memcpy(&srv, *location, sizeof(srv_header));
        printf("Target %d %d %d ", ntohs(srv.priority), ntohs(srv.weight),
         ntohs(srv.port));
      }
      decode_name(&name, buffer, len, ntohs(header.type) == targetType);
      if (ntohs(header.type) == targetType)
        printf("\n");
      break;
    }
  }
  *location += ntohs(header.length);

  return 0;
}

static int decode_dns(char* buffer, int len, int targetType)
{
  HEADER hdr;
  char *loc = buffer + sizeof(HEADER);
  int i = 0, ret;

  if (len < sizeof(HEADER))
    return -1;

  memcpy(&hdr, buffer, sizeof(hdr));
  ret = 0;
  for (i = 0; i < ntohs(hdr.qdcount) && ret == 0; i++)
    ret = decode_question(&loc, buffer, len);

  for (i = 0; i < ntohs(hdr.ancount) && ret == 0; i++)
    ret = decode_rr(&loc, buffer, len, targetType);
  return ret;
}

int queryWithType(const char *prefix, const char *domain, int type)
{
  int ret;

  assert(prefix != NULL);
  if (prefix == NULL)
    return -1;

  if (type == T_SRV)
  {
    char buf[1500]; // TODO:  magic #!
    char *lastDot = strrchr(prefix, '.');

    if (lastDot == NULL || *(lastDot + 1) != '\0')
      ret = res_search(prefix, C_IN, type, buf, sizeof(buf));
    else
      ret = res_query(prefix, C_IN, type, buf, sizeof(buf));

    if (ret == -1 && errno != 0)
      perror("res_query");
    else if (ret > 0)
      decode_dns(buf, ret, type);
  }
  else
    ret = 0; /* no error, I just don't support that option */
  return ret;
}
